/* Z81, a VGA ZX81 emulator for Linux.
 * Copyright (C) 1994 Ian Collier. Z81 changes (C) 1995-6 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <vga.h>
#include <rawkey.h>
#include "z80.h"


unsigned char mem[65536],*helpscrn;
unsigned char *dfile;
unsigned char *memptr[4]={mem,mem+0x4000,mem+0x8000,mem+0xc000};
unsigned char *iptr=mem;
int kports[8];
unsigned long tstates=0,tsmax=(1<<30);
int hires=0,hiscrn;
int help=0;
int topspeed=0,lowref=0;

int memattr[4]={0,2,1,1};

int screen_dirty;
int hsize=256,vsize=192;
volatile int interrupted=0;
unsigned char *vptr;
int forceok=0;
int scrn_freq=2;
int booting=1;
int input_wait=0;

unsigned char chrmap_old[24*32],chrmap[24*32];
unsigned char himap_old[192*32],himap[192*32];

int refresh_screen=1;


void sighandler(a)
int a;
{
if(interrupted<2) interrupted=1;
}


void dontpanic(a)
int a;
{
rawmode_exit();
vga_setmode(TEXT);
exit(1);
}



void screenon()
{
vga_setmode(G320x200x256);
memset(vptr,15,320*200);
refresh_screen=1;
}


void screenoff()
{
vga_setmode(TEXT);
}


char *libdir(char *file)
{
static char buf[1024];
sprintf(buf,"%s/%s",LIBDIR,file);
return(buf);
}


main(argc,argv)
int argc;
char **argv;
{
struct sigaction sa;
struct itimerval itv;
int tmp=1000/50;	/* 50 ints/sec */

vga_init();

loadrom(mem);
hacks();
memset(mem+16384,0xff,49152);
memset(chrmap_old,0xff,768);

loadhelp();
vga_setmode(G320x200x256);
vptr=vga_getgraphmem();
screenon();
rawmode_init();
set_switch_functions(screenoff,screenon);
allow_switch(1);

if(argc==2) scrn_freq=atoi(argv[1]);
if(scrn_freq<1) scrn_freq=1;
if(scrn_freq>50) scrn_freq=50;

sa.sa_handler=dontpanic;
sa.sa_mask=0;
sa.sa_flags=SA_ONESHOT;

sigaction(SIGINT, &sa,NULL);
sigaction(SIGHUP, &sa,NULL);
sigaction(SIGILL, &sa,NULL);
sigaction(SIGTERM,&sa,NULL);
sigaction(SIGQUIT,&sa,NULL);
sigaction(SIGSEGV,&sa,NULL);

sa.sa_handler=sighandler;
sa.sa_mask=0;
sa.sa_flags=SA_RESTART;

sigaction(SIGALRM,&sa,NULL);

itv.it_interval.tv_sec=  tmp/1000;
itv.it_interval.tv_usec=(tmp%1000)*1000;
itv.it_value.tv_sec=itv.it_interval.tv_sec;
itv.it_value.tv_usec=itv.it_interval.tv_usec;
setitimer(ITIMER_REAL,&itv,NULL);

mainloop();

rawmode_exit();
vga_setmode(TEXT);
exit(0);
}


loadrom(x)
unsigned char *x;
{
int i;
FILE *in;

if((in=fopen(libdir("zx81.rom"),"rb"))!=NULL)
  {
  fread(x,1,8192,in);
  memcpy(x+8192,x,8192);
  fclose(in);
  }
else
  {
  printf("Couldn't load ROM. (Have you done 'make install' yet?)\n");
  exit(1);
  }
}


loadhelp()
{
FILE *in;
char buf[128];
int i;

if((in=fopen(libdir("zx81kybd.pgm"),"rb"))!=NULL)
  {
  /* ditch header lines */
  fgets(buf,sizeof(buf),in);
  fgets(buf,sizeof(buf),in);
  fgets(buf,sizeof(buf),in);
  if((helpscrn=malloc(320*200))==NULL)
    {
    printf("Couldn't allocate 64k for help screen.\n");
    exit(1);
    }
  fread(helpscrn,1,320*200,in);
  fclose(in);
  }
else
  {
  printf("Couldn't load help screen 'zx81kybd.pgm'.\n");
  exit(1);
  }

for(i=0;i<320*200;i++)
  helpscrn[i]=(helpscrn[i]*15)/255+16;
}


drawhelp()
{
memcpy(vptr,helpscrn,320*200);
}


hacks()
{
/* remove fast mode stuff which screws it all up */
mem[0x4cc]=mem[0x4cd]=mem[0x4ce]=0;

/* replace kybd routine itself, with ld hl,(lastk):ret */
mem[0x2bb]=0x2a; mem[0x2bc]=0x25; mem[0x2bd]=0x40; mem[0x2be]=0xc9;

/* a 'force F/0' at 66h; ok, as we simulate NMI effects by other means */
mem[0x66]=0xcf; mem[0x67]=0x0e;

/* patch save routine */
mem[0x2fc]=0xed; mem[0x2fd]=0xfd;
mem[0x2fe]=0xc3; mem[0x2ff]=0x07; mem[0x300]=0x02;

/* patch load routine */
mem[0x347]=0xeb;
mem[0x348]=0xed; mem[0x349]=0xfc;
mem[0x34a]=0xc3; mem[0x34b]=0x07; mem[0x34c]=0x02;

/* make fast/slow more traceable (directly modify bit 7 rather than
 * bit 6 of CDFLAGS)
 */
mem[0xf29]=0xbe;
mem[0xf2e]=0xfe;

/* do 'out (0),a' when waiting for input (for fast mode display check) */
mem[0x4ca]=0xd3; mem[0x4cb]=0;			/* out (0),a */
mem[0x4cc]=0xcb; mem[0x4cd]=0x46;		/* loop: bit 0,(hl) */
mem[0x4ce]=0x28; mem[0x4cf]=0xfc;		/* jr z,loop */
mem[0x4d0]=0xd3; mem[0x4d1]=1;			/* out (1),a */
mem[0x4d2]=0;	/* nop, as filler */

/* skip some more weird fast mode stuff */
mem[0x2ec]=0xc9;

/* we need the following crock to support 'PAUSE'... */

mem[0xf3b]=0x29;	/* call 0229h */

mem[0x229]=0x22;	/* ld (04034h),hl:nop */
mem[0x22c]=0;

mem[0x23e]=0x2a; mem[0x23f]=0x25; mem[0x240]=0x40;	/* ld hl,(0403bh) */
mem[0x241]=0x23;					/* inc hl */
mem[0x242]=0x3e; mem[0x243]=0x7f;			/* ld a,07fh */
mem[0x244]=0xa4;					/* and h */
mem[0x245]=0xb5;					/* or l */
mem[0x246]=0xc0;					/* ret nz */

mem[0x247]=0x2a;	/* ld hl,(04034h) */
mem[0x248]=0x34;
mem[0x249]=0x40;

mem[0x24a]=0xc3;	/* jp 022dh */
mem[0x24b]=0x2d;
mem[0x24c]=0x02;

/* nop out ld (04034h),hl */
mem[0x23a]=0; mem[0x23a]=0; mem[0x23a]=0;

/* nop out call to 207h */
mem[0xf41]=0; mem[0xf42]=0; mem[0xf43]=0;
}


unsigned int in(h,l)
int h,l;
{
if(l==0xfe)	/* keyboard */
  switch(h)
    {
    case 0xfe:	return(kports[0]);
    case 0xfd:	return(kports[1]);
    case 0xfb:	return(kports[2]);
    case 0xf7:	return(kports[3]);
    case 0xef:	return(kports[4]);
    case 0xdf:	return(kports[5]);
    case 0xbf:	return(kports[6]);
    case 0x7f:	return(kports[7]);
    default:	return(255);
    }

fprintf(stderr," in l=%02X\n",l);
return(255);
}


unsigned int out(h,l,a)
int h,l,a;
{
if(l==0xfd) store(16443,128|fetch(16443));	/* i.e. set slow mode */
if(l==0x00) input_wait=1;	/* waiting for input */
if(l==0x01) input_wait=0;	/* not waiting for input :-) */

return(0);
}


hireschk(ix)
int ix;
{
refresh_screen=1;

if(ix<0x4000)
  hires=0;
else
  {
  int f,g,h,v;
  
  /* now try to find the hi-res screen.
   * we do this by looking for somewhere in ram where there
   * are 192 suitably-spaced EOL chars. The EOL char could theoretically be
   * a few different things, so we just test that it's the same.
   * this is SLOW, but it doesn't matter; this'll only need to
   * happen once, when the hi-res screen is set up.
   *
   * (um, ok, it's a little different now; I just kept hacking until
   * it worked with all three hi-res programs I tried it on :-))
   */
   
  for(f=0x8000-6400;f>=0x4000;f--)
    {
    v=mem[f+32];
    if((v&64)==0) continue;
    hires=1;
    for(g=0;g<192 && hires;g++)
      {
      if(mem[f+33*g]&64) hires=0;
      if(mem[f+32+33*g]!=v) hires=0;
      }
    if(hires)
      {
      hiscrn=f;
      return(0);
      }
    }
  }
}


hireschki(i)
int i;
{
iptr=mem+(i<<8);
}


/* draw "hi-res" screen */
drawhires()
{
int x,y,c,d,inv,mask,a,ofs;
unsigned char *tmp,*ptr;

ptr=mem+hiscrn;

ofs=0;
for(y=0;y<192;y++,ptr++)
  {
  for(x=0;x<32;x++,ptr++,ofs++)
    {
    c=*ptr;
    
    if((himap[ofs]=c)!=himap_old[ofs] || refresh_screen)
      {
      inv=(c&128); c&=63;
      
      tmp=vptr+4*320+32+y*320+x*8;
      d=iptr[c*8]; if(inv) d^=255;
      mask=128;
      for(a=0;a<8;a++,mask>>=1)
        *tmp++=(d&mask)?0:15;
      }
    }
  }

/* now, copy new to old for next time */
memcpy(himap_old,himap,32*192);
refresh_screen=0;
}


/* redraw the screen */
update_scrn()
{
static int count=0;
int x,y,a,b,c,d,inv,x2,mask;
unsigned char *ptr,*cptr,*tmp;
int fast=(!(fetch(16443)&128) && !input_wait);
int pasteol;
int ofs;

/* only do it every 1/25th */
count++;
if(count<(lowref?25:scrn_freq)) return(0); else count=0;

/* only if not showing help screen */
if(help) return(0);

if(hires && !fast)
  {
  drawhires();
  return(0);
  }

ptr=mem+fetch2(16396);	/* D_FILE */
/* don't bother if it's junk */
if(ptr-mem<0 || ptr-mem>0xf000) return(0);
ptr++;	/* skip first HALT */

cptr=mem+0x1e00;

ofs=0;
for(y=0;y<24;y++,ptr++)
  {
  pasteol=0;
  for(x=0;x<32;x++,ptr++,ofs++)
    {
    if(pasteol || (*ptr)==0x76)
      {
      pasteol=1;
      c=0;	/* space */
      ptr--;
      }
    else
      c=*ptr;
    
    if((chrmap[ofs]=c)!=chrmap_old[ofs] || refresh_screen)
      {
      inv=(c&128); c&=63;
      
      for(b=0;b<8;b++)
        {
        tmp=vptr+4*320+32+(y*8+b)*320+x*8;
        d=cptr[c*8+b]; if(inv) d^=255;
        mask=128;
        for(a=0;a<8;a++,mask>>=1)
          *tmp++=(d&mask)?0:15;
        }
      }
    }
  }

/* now, copy new to old for next time */
memcpy(chrmap_old,chrmap,768);
refresh_screen=0;
}


update_kybd()
{
int y,b;
int lastk0=0,lastk1=0;	/* values for lastk and lastk+1 */
static int oldlk0=0,oldlk1=0;
int f;
unsigned short tmp;

/* XXX want to support DEL etc. */

/* bump frames */
f=(fetch2(16436)&32767);
if(f>0) f--;
store2(16436,(fetch2(16436)&32768)|f);

/* set max T-states per 1/50th to get the right speed */
if(topspeed)
  tsmax=(1<<30);
else
  if(fetch(16443)&128)
    {
    /* slow mode - 0.8 MHz, 16000 T-states per 1/50th */
    tsmax=16000;
    }
  else
    {
    /* fast mode - 3.2 MHz, 64000 T-states per 1/50th */
    tsmax=64000;
    }

for(y=0;y<5;y++) scan_keyboard();

if(is_key_pressed(FUNC_KEY(1)))
  {
  if(help)
    memset(vptr,15,320*200),help=0,refresh_screen=1;
  else
    drawhelp(),help=1;
  while(is_key_pressed(FUNC_KEY(1))) { usleep(20000); scan_keyboard(); }
  }

if(is_key_pressed(ESCAPE_KEY))
  {
  interrupted=2;	/* will cause a reset */
  memset(mem+16384,0xff,49152);
  iptr=mem;
  tsmax=(1<<30);
  hires=0; refresh_screen=1;
  booting=1; input_wait=0;
  }

if(is_key_pressed(FUNC_KEY(8)))
  {
  lowref=!lowref;
  if(lowref)
    vga_setpalette(15,63,63,50);
  else
    vga_setpalette(15,63,63,63);
  while(is_key_pressed(FUNC_KEY(8))) { usleep(20000); scan_keyboard(); }
  }

if(is_key_pressed(FUNC_KEY(9)))
  {
  topspeed=!topspeed;
  if(topspeed)
    vga_setpalette(0,0,0,60);
  else
    vga_setpalette(0,0,0,0);
  while(is_key_pressed(FUNC_KEY(9))) { usleep(20000); scan_keyboard(); }
  }

if(is_key_pressed(FUNC_KEY(10)))
  {
  while(is_key_pressed(FUNC_KEY(10))) { usleep(20000); scan_keyboard(); }
  dontpanic();	/* F10 = quit */
  }

/* ugly, but there's no pleasant way to do this */
/* OTOH, there's definitely a better way than this pile of cack :-) */

for(y=0;y<8;y++)		/* 8 half-rows */
  {
  b=0;	/* we set bits in b as appropriate */
  switch(y)	/* below comments given in order b1->b5 */
    {
    /* left-hand side */
    case 0:	/* sft,z,x,c,v */
      if(is_key_pressed(LEFT_SHIFT) || is_key_pressed(RIGHT_SHIFT) ||
         is_key_pressed(LEFT_CTRL))
        b|=1;	/* special case - shift is bit 0 */
      if(is_key_pressed(scancode_trans('z'))) b|=4;
      if(is_key_pressed(scancode_trans('x'))) b|=8;
      if(is_key_pressed(scancode_trans('c'))) b|=16;
      if(is_key_pressed(scancode_trans('v'))) b|=32;
      break;
    case 1:	/* a,s,d,f,g */
      if(is_key_pressed(scancode_trans('a'))) b|=2;
      if(is_key_pressed(scancode_trans('s'))) b|=4;
      if(is_key_pressed(scancode_trans('d'))) b|=8;
      if(is_key_pressed(scancode_trans('f'))) b|=16;
      if(is_key_pressed(scancode_trans('g'))) b|=32;
      break;
    case 2:	/* q,w,e,r,t */
      if(is_key_pressed(scancode_trans('q'))) b|=2;
      if(is_key_pressed(scancode_trans('w'))) b|=4;
      if(is_key_pressed(scancode_trans('e'))) b|=8;
      if(is_key_pressed(scancode_trans('r'))) b|=16;
      if(is_key_pressed(scancode_trans('t'))) b|=32;
      break;
    case 3:	/* 1,2,3,4,5 */
      if(is_key_pressed(scancode_trans('1'))) b|=2;
      if(is_key_pressed(scancode_trans('2'))) b|=4;
      if(is_key_pressed(scancode_trans('3'))) b|=8;
      if(is_key_pressed(scancode_trans('4'))) b|=16;
      if(is_key_pressed(scancode_trans('5'))) b|=32;
      break;

    /* right-hand side */
    case 4:	/* 0,9,8,7,6 */
      if(is_key_pressed(scancode_trans('0'))) b|=2;
      if(is_key_pressed(scancode_trans('9'))) b|=4;
      if(is_key_pressed(scancode_trans('8'))) b|=8;
      if(is_key_pressed(scancode_trans('7'))) b|=16;
      if(is_key_pressed(scancode_trans('6'))) b|=32;
      break;
    case 5:	/* p,o,i,u,y */
      if(is_key_pressed(scancode_trans('p'))) b|=2;
      if(is_key_pressed(scancode_trans('o'))) b|=4;
      if(is_key_pressed(scancode_trans('i'))) b|=8;
      if(is_key_pressed(scancode_trans('u'))) b|=16;
      if(is_key_pressed(scancode_trans('y'))) b|=32;
      break;
    case 6:	/* ent,l,k,j,h */
      if(is_key_pressed(ENTER_KEY)) b|=2;
      if(is_key_pressed(scancode_trans('l'))) b|=4;
      if(is_key_pressed(scancode_trans('k'))) b|=8;
      if(is_key_pressed(scancode_trans('j'))) b|=16;
      if(is_key_pressed(scancode_trans('h'))) b|=32;
      break;
    case 7:	/* spc,dot,m,n,b */
      if(is_key_pressed(scancode_trans(' '))) b|=2;
      if(is_key_pressed(scancode_trans('.'))) b|=4;
      if(is_key_pressed(scancode_trans('m'))) b|=8;
      if(is_key_pressed(scancode_trans('n'))) b|=16;
      if(is_key_pressed(scancode_trans('b'))) b|=32;
      break;
    }
  
  /* now make the byte the ZX81 will want to see at the given port,
   * which is a bit of a pain given the way I've done all this :-)
   */
  kports[y]=((y==0)?(((b&1)|((b>>1)&0xfc))^63):((b>>1)^63));
  kports[y]|=0xe0;	/* some things need top bits to be 1 */
  
  /* and contribute to lastk if key was pressed */
  if(b)
    {
    /* set y bit in lastk0 if not shift */
    if(b>1) lastk0|=(1<<y);
    
    /* set pos-in-half-row bit(s) in lastk1 */
    lastk1|=b;
    }
  }

/* see if this is different to prev. lastk[01] */
store(16443,(fetch(16443)&0xfe)|
	((((lastk1&0xfe)|lastk0) && (oldlk0!=lastk0 || oldlk1!=lastk1))?1:0));

oldlk0=lastk0; oldlk1=lastk1;

/* fix lastk values */
lastk0^=255; lastk1^=255;

/* put lastk stuff in place */
store(16421,lastk0);
store(16422,lastk1);

/* and FWIW, set MARGIN */
store(16424,55);
}


/* the ZX81 char is used to index into this, to give the ascii.
 * awkward chars are mapped to '_' (underscore), and the ZX81's
 * all-caps is converted to lowercase.
 * WARNING: this only covers 0<=char<=63!
 */
char zx2ascii[64]={
/* 000-009 */ ' ', '_', '_', '_', '_', '_', '_', '_', '_', '_', 
/* 010-019 */ '_', '\'','#', '$', ':', '?', '(', ')', '>', '<', 
/* 020-029 */ '=', '+', '-', '*', '/', ';', ',', '.', '0', '1', 
/* 030-039 */ '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 
/* 040-049 */ 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 
/* 050-059 */ 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 
/* 060-063 */ 'w', 'x', 'y', 'z'
};


save_p(a)
int a;
{
static char fname[128];
unsigned char *ptr=mem+a,*dptr=fname;
FILE *out;

/* so the filename is at ptr, in the ZX81 char set, with the last char
 * having bit 7 set. First, get that name in ASCII.
 */

memset(fname,0,sizeof(fname));
do
  *dptr++=zx2ascii[(*ptr)&63];
while((*ptr++)<127);

/* add '.p' */
strcat(fname,".p");

/* now open the file */
if((out=fopen(fname,"wb"))==NULL)
  {
  forceok=1;
  return(0);
  }

/* work out how much to write, and write it.
 * we need to write from 0x4009 to E_LINE inclusive.
 */
fwrite(mem+0x4009,1,fetch2(16404)-0x4009+1,out);

fclose(out);
}


load_p(a)
int a;
{
static char fname[128];
unsigned char *ptr=mem+(a&32767),*dptr=fname;
FILE *in;

if(a&32768) 	/* they did LOAD "" */
  {
  forceok=1;
  return(0);
  }

/* so the filename is at ptr, in the ZX81 char set, with the last char
 * having bit 7 set. First, get that name in ASCII.
 */

/* test for Xtender-style LOAD " STOP " to quit */
if(*ptr==227) dontpanic();

memset(fname,0,sizeof(fname));
do
  *dptr++=zx2ascii[(*ptr)&63];
while((*ptr++)<127);

/* add '.p' */
strcat(fname,".p");

/* now open the file */
if((in=fopen(fname,"rb"))==NULL)
  {
  forceok=1;
  return(0);
  }

/* just read it all */
fread(mem+0x4009,1,16384,in);

fclose(in);
}


fix_tstates()
{
tstates=0;
pause();
}


do_interrupt()
{
if(booting)
  if(input_wait)
    booting=0;
  else
    return(0);

update_scrn();
update_kybd();
if(interrupted<2) interrupted=0;
}
